# -*- coding: utf-8 -*-
"""This module provides support for breaking substitution ciphers.

This includes generating quadgrams from a text corpus, scoring a plaintext and
breaking substitution ciphers.
"""

import math
import json
import sys
import random
import time

from subbreaker.key import Key, AlphabetInvalid


class BreakerInfo(object):
    """Class representing various information of the quadgrams for a given language

    :ivar str alphabet: text representation of the alphabet
    :ivar int nbr_quadgrams: the number of quadgrams considered for the generated
        quadgram file. This value may be lower than the corpus size as only characters
        from the alphabet are considered for this value
    :ivar str most_frequent_quadgram: the most often occurring same sequence of four
        characters within the corpus used to generate the quadgram file. For English,
        the expected value here is "tion".
    :ivar float average_fitness: the expected fitness of a text if all characters are
        generated randomly with the same probability.
    :ivar float max_fitness: the fitness for the most frequent quadgram
    """

    def __init__(
        self,
        alphabet=None,
        nbr_quadgrams=None,
        most_frequent_quadgram=None,
        average_fitness=None,
        max_fitness=None,
    ):
        self.alphabet = alphabet
        self.nbr_quadgrams = nbr_quadgrams
        self.most_frequent_quadgram = most_frequent_quadgram
        self.average_fitness = average_fitness
        self.max_fitness = max_fitness


class BreakerResult(object):
    """Class representing the result for breaking a substitution cipher

    :ivar str ciphertext: the original ciphertext
    :ivar str plaintext: the resulting plaintext using the found key
    :ivar str key: the best key found by the breaker
    :ivar str alphabet: the alphabet used to break the cipher
    :ivar float fitness: the fitness of the resulting plaintext
    :ivar int nbr_keys: the number of keys tried by the breaker
    :ivar nbr_rounds: the number of hill climbings performed, starting with a random
        key
    :ivar float keys_per_second: the number of keys tried per second
    :ivar float seconds: the time in seconds used to break the cipher
    """

    def __init__(
        self,
        ciphertext=None,
        plaintext=None,
        key=None,
        alphabet=None,
        fitness=0,
        nbr_keys=0,
        nbr_rounds=0,
        keys_per_second=0,
        seconds=0,
    ):
        """Instantiation method
        """
        self.ciphertext = ciphertext
        self.plaintext = plaintext
        self.key = key
        self.alphabet = alphabet
        self.fitness = fitness
        self.nbr_keys = nbr_keys
        self.nbr_rounds = nbr_rounds
        self.keys_per_second = keys_per_second
        self.seconds = seconds

    def __str__(self):
        return "key = {}".format(self.key)


class Breaker(object):
    """Class to represent the breaker implementation based on quadgrams

    :ivar info: a :class:`BreakerInfo` object
    :type info: :class:`BreakerInfo` object
    :ivar key: a :class:`Key` object with the key found when breaking a cipher
    :type key: :class:`Key`
    :param quadgram_fh: file handle (i.e., a read()-supporting file like object)
        to read the quadgrams from.
    :type quadgram_fh: file handle
    """

    _DEFAULT_ALPHABET = "abcdefghijklmnopqrstuvwxyz"

    def __init__(self, quadgram_fh):
        """Init the instance
        """
        obj = json.load(quadgram_fh)
        self._alphabet = obj["alphabet"]
        self._alphabet_len = len(self._alphabet)
        self._quadgrams = obj["quadgrams"]
        self.info = BreakerInfo(
            alphabet=obj["alphabet"],
            nbr_quadgrams=obj["nbr_quadgrams"],
            most_frequent_quadgram=obj["most_frequent_quadgram"],
            average_fitness=obj["average_fitness"] / 10,
            max_fitness=obj["max_fitness"] / 10,
        )
        self.key = None

    @staticmethod
    def _file_iterator(file_fh, alphabet):
        """Implements an iterator for a given file based text file

        The iterator will yield all charcters of the text file which are present in
        the alphabet, all other characters will be skipped.

        :param file_fh: the file handle (i.e., a read()-supporting file like object)
            of the text file
        :type file_fh: file handle
        :param str alphabet: the alphabet to apply with this text file.
        :return: an iterator which iterates over all characters of the text file
            which are present in the alphabet.
        """
        trans = {val: key for key, val in enumerate(alphabet.lower())}
        for line in file_fh:
            line = line.lower()
            for char in line:
                val = trans.get(char)
                if val is not None:
                    yield val

    @staticmethod
    def _text_iterator(txt, alphabet):
        """Implements an iterator for a given text string

        The iterator will yield all characters of the text string which are present in
        the alphabet, all other characters will be skipped.

        :param str txt: the text string to process
        :param str alphabet: the alphabet to apply with this text string
        :return: an iterator which iterates over all characters of the text string
            which are present in the alphabet.
        """
        trans = {val: key for key, val in enumerate(alphabet.lower())}
        for char in txt.lower():
            val = trans.get(char)
            if val is not None:
                yield val

    @staticmethod
    def generate_quadgrams(corpus_fh, quadgram_fh, alphabet=_DEFAULT_ALPHABET):
        """Static method to generate quadgrams from a text file

        Based on the text file the quadgrams will be serialized in JSON format into
        the file "quadgram_file".

        The quadgrams will be generated by the following steps:

        - the number of occurrences of the quadgram within the text file is counted
          per quadgram
        - if a quadgram does not occur in the text file (i.e. the counted value
          is zero), it will be set to 1/10th of the least often occurring quadgram. This
          is done to allow calculating the logarithm and avoiding "math domain errors"
        - the logarithm of the number of occurrences for each quadgram is calculated
        - the results are normalized (which is not strictly required, but it provides
          fitness values which can be somehow interpreted by humans)

            - to avoid negative values
            - to have a reference point in such a way that value = 100 corresponds to
              "typical occurrences" of the quadgram in the given language. Values below
              100 indicate "less often occurring quadgrams" (the lower the value the
              lesser it will statistically occur), while values above 100 indicate that
              these quadgrams are occurring more often then average in the given
              language. Is that clear enough?

        The quadgrams as well as the used alphabet is then serialized into a JSON
        structure and store in the file "quadgram_file".

        :param corpus_fh: the file handle (i.e., a read()-supporting file like object)
            of the text corpus file to process
        :type corpus_fh: file handle
        :param str quadgram_fh: the file handle (i.e., a write()-supporting file like
            object) where the quadgrams will be serialized to
        :type quadgram_fh: file handle
        :param str alphabet: the alphabet to apply with this text file.
        :raises: ValueError - if the alphabet has more than 32 characters
        """
        alphabet = Key.check_alphabet(alphabet)
        if len(alphabet) > 32:
            raise AlphabetInvalid("Alphabet must have less or equal than 32 characters")
        iterator = Breaker._file_iterator(corpus_fh, alphabet)
        quadgram_val = iterator.__next__()
        quadgram_val = (quadgram_val << 5) + iterator.__next__()
        quadgram_val = (quadgram_val << 5) + iterator.__next__()
        quadgrams = [0 for cntr in range(32 * 32 * 32 * 32)]
        for numerical_char in iterator:
            quadgram_val = ((quadgram_val & 0x7FFF) << 5) + numerical_char
            quadgrams[quadgram_val] += 1

        quadgram_sum = sum(quadgrams)
        quadgram_min = 10000000
        for val in quadgrams:
            if val:
                quadgram_min = min(quadgram_min, val)
        offset = math.log(quadgram_min / 10 / quadgram_sum)

        norm = 0
        for idx, val in enumerate(quadgrams):
            if val:
                prop = val / quadgram_sum
                new_val = math.log(prop) - offset
                quadgrams[idx] = new_val
                norm += prop * new_val

        for idx, val in enumerate(quadgrams):
            quadgrams[idx] = round(quadgrams[idx] / norm * 1000)

        # Just for curiosity: determine the most frequent quadgram
        max_idx = quadgrams.index(max(quadgrams))
        max_val = quadgrams[max_idx]
        # now construct the ASCII representation from the index
        max_chars = []
        index = max_idx
        for _ in range(4):
            max_chars = [alphabet[index & 0x1F]] + max_chars
            index >>= 5

        json.dump(
            {
                "alphabet": alphabet,
                "nbr_quadgrams": quadgram_sum,
                "most_frequent_quadgram": "".join(max_chars),
                "max_fitness": max_val,
                "average_fitness": sum(quadgrams) / (len(alphabet) ** 4),
                "quadgrams": quadgrams,
            },
            quadgram_fh,
            indent=0,
        )

    def _calc_fitness(self, iterator):
        """Calculate the fitness from the characters provided by the iterator

        :param iterator: iterator which provides the characters relevant for
            calculating the fitness
        :param type: iterator object
        :return: the fitness of the text. A value close to 100 means, the
            text is probably in the same language than the language used to generate
            the quadgrams. The more the value differs from 100, the lesser the
            probability that the examined text corresponds to the quadgram language.
            Lower values indicate more random text, while values significantly
            greater than 100 indicate (nonsense) text with too much frequently used
            quadgrams (e.g., ``tionioningatheling``).
        :rtype: float
        :raises: ValueError
        """
        try:
            quadgram_val = iterator.__next__()
            quadgram_val = (quadgram_val << 5) + iterator.__next__()
            quadgram_val = (quadgram_val << 5) + iterator.__next__()
        except StopIteration:
            raise ValueError(
                "More than three characters from the given alphabet are required"
            )

        fitness = 0
        nbr_quadgrams = 0
        quadgrams = self._quadgrams
        for numerical_char in iterator:
            quadgram_val = ((quadgram_val & 0x7FFF) << 5) + numerical_char
            fitness += quadgrams[quadgram_val]
            nbr_quadgrams += 1
        if nbr_quadgrams == 0:
            raise ValueError(
                "More than three characters from the given alphabet are required"
            )
        return fitness / nbr_quadgrams / 10

    def calc_fitness_file(self, cleartext_fh=sys.stdin):
        """Method to calculate the fitness of the given file contents

        :param cleartext_fh: the file handle (i.e., a read()-supporting file like
            object) from which the fitness will be calculated
        :type cleartext_fh: file handle
        :return: the fitness of the text. A value close to 100 means, the
            text is probably in the same language than the language used to generate
            the quadgrams. The more the value differs from 100, the lesser the
            probability that the examined text corresponds to the quadgram language.
            Lower values indicate more random text, while values significantly
            greater than 100 indicate (nonsense) text with too much frequently used
            quadgrams (e.g., ``tionioningatheling``).
        :rtype: float
        """
        return self._calc_fitness(Breaker._file_iterator(cleartext_fh, self._alphabet))

    def calc_fitness(self, txt):
        """Method to calculate the fitness for the given text string

        :param str txt: the text string for which the fitness shall be determined
        :return: the fitness of the text. A value close to 100 means, the
            text is probably in the same language as the language used to generate
            the quadgrams. The more the value differs from 100, the lesser the
            probability that the examined text corresponds to the quadgram language.
            Lower values indicate more random text, while values significantly
            greater than 100 indicate (nonsense) text with too much frequently used
            quadgrams (e.g., ``tionioningatheling``).
        :rtype: float
        """
        return self._calc_fitness(Breaker._text_iterator(txt, self._alphabet))

    def _hill_climbing(self, key, cipher_bin, char_positions):
        """Basic hill climbing function

        Starting from the given key two charters are swapped, hoping the mutated key
        provides a better result than the original key. This is done for all possible
        pairs of characters. If a better key is found, the mutation process starts
        over again. If no better key is found, we reached a (local) maximum and the
        fitness and the number of keys is returned.

        :param key: the key to start with. The key is represented as a list where each
            character is represented by an integer 0..25 (for the default alphabet).
            0 corresponds to the first character of the alphabet, 1 to the second,
            etc.
        :type key: list(int)
        :param cipher_bin: the ciphertext in binary representation, i.e. each
            character present in the alphabet is represented as an integer.
        :type cipher_bin: list(int)
        :param char_positions: a list for each character of the alphabet indicating
            at which positions it is occurring. The positions are stored in a list.
            Used to quickly update the plaintext when two characters of the key are
            swapped.
        :type char_positions: list(list(int))
        which for each character holds a list where in
            the ciphertext it occurs.
        :return: tuple of the max_fitness and the number of keys evaluated
        :rtype: tuple
        """
        plaintext = [key.index(idx) for idx in cipher_bin]
        quadgram = self._quadgrams
        key_len = self._alphabet_len
        nbr_keys = 0
        max_fitness = 0
        better_key = True
        while better_key:
            better_key = False
            for idx1 in range(key_len - 1):
                for idx2 in range(idx1 + 1, key_len):
                    ch1 = key[idx1]
                    ch2 = key[idx2]
                    for idx in char_positions[ch1]:
                        plaintext[idx] = idx2
                    for idx in char_positions[ch2]:
                        plaintext[idx] = idx1
                    nbr_keys += 1
                    tmp_fitness = 0
                    quad_idx = (plaintext[0] << 10) + (plaintext[1] << 5) + plaintext[2]
                    for char in plaintext[3:]:
                        quad_idx = ((quad_idx & 0x7FFF) << 5) + char
                        tmp_fitness += quadgram[quad_idx]
                    if tmp_fitness > max_fitness:
                        max_fitness = tmp_fitness
                        better_key = True
                        key[idx1] = ch2
                        key[idx2] = ch1
                    else:
                        for idx in char_positions[ch1]:
                            plaintext[idx] = idx1
                        for idx in char_positions[ch2]:
                            plaintext[idx] = idx2
        return max_fitness, nbr_keys

    def break_cipher(self, ciphertext, max_rounds=10000, consolidate=3):
        """Breaks a given cipher text

        :param str ciphertext: the ciphertext to break
        :param int max_rounds: the maximum number of hill climbing rounds to execute
        :param int consolidate: the number of times the best local maximum must be
            reached before it is considered as the overall best solution
        :return: an BreakerResult object
        :rtype: :class:`BreakerResult`
        """
        if not (1 <= max_rounds <= 10000):
            raise ValueError("maximum number of rounds not in the valid range 1..10000")
        if not (1 <= consolidate <= 30):
            raise ValueError("consolidate parameter out of valid range 1..30")
        start_time = time.time()
        nbr_keys = 0
        cipher_bin = [
            char for char in Breaker._text_iterator(ciphertext, self._alphabet)
        ]
        if len(cipher_bin) < 4:
            raise ValueError("ciphertext is too short")

        char_positions = []
        for idx in range(len(self._alphabet)):
            char_positions.append([i for i, x in enumerate(cipher_bin) if x == idx])

        key_len = len(self._alphabet)
        local_maximum, local_maximum_hit = 0, 1
        key = [idx for idx in range(key_len)]
        best_key = key.copy()
        for round_cntr in range(max_rounds):
            random.shuffle(key)
            fitness, tmp_nbr_keys = self._hill_climbing(key, cipher_bin, char_positions)
            nbr_keys += tmp_nbr_keys
            if fitness > local_maximum:
                local_maximum = fitness
                local_maximum_hit = 1
                best_key = key.copy()
            elif fitness == local_maximum:
                local_maximum_hit += 1
                if local_maximum_hit == consolidate:
                    break
        key_str = "".join([self._alphabet[x] for x in best_key])
        self.key = Key(key_str, alphabet=self._alphabet)
        seconds = time.time() - start_time
        return BreakerResult(
            ciphertext=ciphertext,
            plaintext=self.key.decode(ciphertext),
            key=key_str,
            alphabet=self._alphabet,
            fitness=local_maximum / (len(cipher_bin) - 3) / 10,
            nbr_keys=nbr_keys,
            nbr_rounds=round_cntr,
            keys_per_second=round(nbr_keys / seconds, 3),
            seconds=seconds,
        )



# -*- coding: utf-8 -*-
import pytest
import pathlib


@pytest.fixture
def fixturefiles_dir():
    return pathlib.Path(__file__).resolve().parent / "fixturefiles"


@pytest.fixture
def hello_txt(fixturefiles_dir):
    return str(fixturefiles_dir / "Hello.txt")


@pytest.fixture
def ciphertext_file_name(fixturefiles_dir):
    return str(fixturefiles_dir / "fixed_cipher.txt")


@pytest.fixture
def fixed_ciphertext(ciphertext_file_name):
    with open(ciphertext_file_name) as fh:
        ciphertext = fh.read()
    return ciphertext


@pytest.fixture
def quadgram_file():
    return (
        pathlib.Path(__file__).resolve().parents[1]
        / "subbreaker"
        / "quadgram"
        / "EN.json"
    )


@pytest.fixture
def quadgram_fh(quadgram_file):
    with open(str(quadgram_file)) as fh:
        yield fh



# -*- coding: utf-8 -*-
"""Module containing the CLI implementation
"""
import argparse
import sys
import random
import pathlib
from contextlib import contextmanager

from subbreaker.key import Key, AlphabetInvalid, KeyInvalid
from subbreaker.breaker import Breaker

quadgram_dir = pathlib.Path(__file__).resolve().parent / "quadgram"


@contextmanager
def open_file(file_name, mode="r"):
    """Unify handling of stdin/stdout and files to read from resp. write to

    This is a helper function which can be used with context managers.

    :param file_name: The name of the file (full path) to read from. If None is given,
        stdin will be used
    :type file_name: str or None
    :param str mode: The mode used to open the file, defaults to "r"
    """
    default = sys.stdin if "r" in mode else sys.stdout
    fh = default if file_name is None else open(file_name, mode)
    try:
        yield fh
    finally:
        if file_name is not None:
            fh.close()


def consistency_check(args, alphabet, key):
    """Checks the alphabet and the key for consistency and normalizes both to lower case

    :param object args: object containing the converted arguments
    :param str alphabet: the untrusted alphabet to check
    :param str key: the untrusted key to check
    :return: a tuple of the validated alphabet and key. Both are returned in lower case.
    :rtype: tuple(str, str)
    """
    try:
        alphabet = Key.check_alphabet(alphabet)
        key = Key.check_key(key, alphabet)
    except (AlphabetInvalid, KeyInvalid) as exc:
        args.subparser.error(str(exc))
    return alphabet, key


def build_key_from_keyword(args):
    """In case a keyword was given, determine the corresponding key

    :param object args: object containing the converted arguments
    """
    alpha_list = list(args.alphabet.lower())
    kw_list = []
    for char in args.keyword.lower():
        if char not in kw_list:
            kw_list.append(char)
            alpha_list.remove(char)
    return "".join(kw_list + alpha_list)


def command_version(args):
    """Prints the version.

    :param object args: object containing the converted arguments
    """
    if sys.version_info >= (3,8):
        from importlib.metadata import version, PackageNotFoundError
    else:
        from importlib_metadata import version, PackageNotFoundError
    try:
        print(version("subbreaker"))
    except PackageNotFoundError:
        # package is not installed
        print("Cannot determine the version")

def command_fitness(args):
    """Calculates the fitness for a given language and text

    :param object args: object containing the converted arguments
    """
    quadgram_file = pathlib.Path(quadgram_dir) / (args.lang + ".json")
    with open(str(quadgram_file)) as fh:
        breaker = Breaker(fh)
    if args.text is not None:
        fitness = breaker.calc_fitness(args.text)
    else:
        with open_file(args.in_file) as fh:
            fitness = breaker.calc_fitness_file(fh)
    print(round(fitness, 2))


def command_break(args):
    """Breaks a substitution cipher

    :param object args: object containing the converted arguments
    """
    if not (1 <= args.max_tries <= 10000):
        args.subparser.error("--max-tries must be in the range 1..10000")
    if not (1 <= args.consolidate <= 30):
        args.subparser.error("--consolidate must be in the range 1..30")
    quadgram_file = pathlib.Path(quadgram_dir) / (args.lang + ".json")
    with open(str(quadgram_file)) as fh:
        breaker = Breaker(fh)
    if args.text is not None:
        text = args.text
    else:
        with open_file(args.in_file) as fh:
            text = fh.read()
    result = breaker.break_cipher(
        text, max_rounds=args.max_tries, consolidate=args.consolidate
    )
    print("Alphabet: {}".format(result.alphabet))
    print("Key:      {}".format(result.key))
    print("Fitness: {}".format(round(result.fitness, 2)))
    print("Nbr keys tried: {}".format(result.nbr_keys))
    print("Keys per second: {}".format(round(result.keys_per_second)))
    print("Execution time (seconds): {}".format(round(result.seconds, 3)))
    print("Plaintext:")
    print(result.plaintext)


def command_info(args):
    quadgram_file = pathlib.Path(quadgram_dir) / (args.lang + ".json")
    with open(str(quadgram_file)) as fh:
        breaker = Breaker(fh)
    print("Quadgram file: {}".format(str(quadgram_file.resolve())))
    print("Alphabet: {}".format(breaker.info.alphabet))
    print("Length of alphabet: {}".format(len(breaker.info.alphabet)))
    print("Number of quagrams: {}".format(breaker.info.nbr_quadgrams))
    print("Most frequent quadgram: {}".format(breaker.info.most_frequent_quadgram))
    print("Fitness for most frequent quadgram: {}".format(breaker.info.max_fitness))
    print("Fitness for random text: {}".format(round(breaker.info.average_fitness, 2)))


def command_decode(args):
    """Decodes a substitution cipher with a given key

    :param object args: object containing the arguments from the command line
    """
    if args.key is not None:
        key_str = args.key
    else:
        # keyword given
        key_str = build_key_from_keyword(args)

    alphabet, key_str = consistency_check(args, args.alphabet, key_str)

    key = Key(key_str, alphabet)
    if args.text is not None:
        print(key.decode(args.text))
    else:
        with open_file(args.in_file) as ciphertext_fh, open_file(
            args.out_file, "w"
        ) as plaintext_fh:
            key.decode_file(ciphertext_fh, plaintext_fh)


def command_encode(args):
    """Encodes a substitution cipher with a given key

    :param object args: object containing the arguments from the command line
    """
    if args.random:
        key_list = list(args.alphabet)
        random.shuffle(key_list)
        key_str = "".join(key_list)
    elif args.key is not None:
        key_str = args.key
    else:
        # keyword given
        key_str = build_key_from_keyword(args)

    alphabet, key_str = consistency_check(args, args.alphabet, key_str)

    if args.random:
        print(key_str, file=sys.stderr)
    key = Key(key_str, args.alphabet)
    if args.text is not None:
        print(key.encode(args.text))
    else:
        with open_file(args.in_file) as plaintext_fh, open_file(
            args.out_file, "w"
        ) as ciphertext_fh:
            key.encode_file(plaintext_fh, ciphertext_fh)


def command_quadgrams(args):
    """Creates the quadgrams

    :param object args: object containing the converted arguments
    """
    with open_file(args.corpus) as inp_fh, open_file(args.quadgrams, "w") as outp_fh:
        Breaker.generate_quadgrams(inp_fh, outp_fh, args.alphabet)


def args_add_text_or_in_file(parser, arg_name):
    """Adds the mutual exclusive arguments text and input-file to the parser

    :param parser: parser object to extend with the parameters
    :type parser: object returned by add_parser
    :param str arg_name: the name of the argument, used for the file,
        e.g., "--ciphertext"
    """
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--text",
        metavar="<string>",
        help=(
            "string containing the input text. Note, that line breaks and blanks "
            "might require shell escaping."
        ),
    )
    group.add_argument(
        arg_name,
        dest="in_file",
        metavar="<path>",
        help=(
            "name of the file containing the input text. If neither --text nor "
            "{} is given, the text will be read from STDIN.".format(arg_name)
        ),
    )


def args_add_out_file(parser, arg_name):
    """Adds the argument for the output file to the parser

    :param parser: parser object to extend with the parameters
    :type parser: object returned by add_parser
    :param str arg_name: the name of the argument, used for the file,
        e.g., "--ciphertext"
    """
    parser.add_argument(
        arg_name,
        dest="out_file",
        metavar="<path>",
        help=(
            "name of the file the output is written to. If it is not given, the "
            "output is printed to STDOUT."
        ),
    )


def args_add_key_alternatives(parser, random=False):
    """Add the parameter "--key" to the argparser

    :param parser: parsers object to extend with the key parameter
    :type parsers: object returned by add_subparsers
    :param bool required: indicates if --key is a mandatory parameter
    """
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--key",
        metavar="<string>",
        help=(
            "key containing all characters from the alphabet. The key can be specified "
            "case insensitive, may not contain any non-alphabetical characters, "
            "and every character of the alphabet must be present exactly once."
        ),
        default=None,
    )
    group.add_argument(
        "--keyword",
        metavar="<string>",
        help=(
            "a case-insensitive keyword used to build the key. The key is created by "
            "writing out the keyword, removing repeated letters in it, then writing "
            "all the remaining letters in the alphabet in the usual order. E.g., the "
            'keyword "ZEBRAS" leads to the key "zebrascdfghijklmnopqtuvwxy".'
        ),
        default=None,
    )
    if random:
        group.add_argument(
            "--random",
            action="store_true",
            default=False,
            help=(
                "use a random key by shuffling the alphabet. The key is printed to "
                "STDERR."
            ),
        )


def args_add_alphabet(parser):
    """Add the parameter "--alphabet" to the argparser

    :param parser: parsers object to extend with the key parameter
    :type parsers: object returned by add_subparsers
    """
    parser.add_argument(
        "--alphabet",
        metavar="<string>",
        help=(
            "a string of characters which build the alphabet. Lower and upper "
            "characters are treated the same. By default, 'abcdefghijklmnopqrstuvwxyz' "
            "is used, but any character is allowed (including the blank and e.g., "
            "umlauts). The length of the alphabet may not exceed 32 characters."
        ),
        default="abcdefghijklmnopqrstuvwxyz",
    )


def args_add_language(parser):
    """Add the parameter "--lang" to the argparser

    :param parser: parsers object to extend with the key parameter
    :type parsers: object returned by add_subparsers
    """
    languages = [file_name.stem for file_name in quadgram_dir.glob("*.json")]
    lang_str = "{" + "|".join(languages) + "}"
    parser.add_argument(
        "--lang",
        metavar=lang_str,
        help="language of the text. The default is EN for English.",
        type=str,
        default="EN",
    )


def args_break(subparsers):
    """Defines the arguments for the subcommand "break"

    :param subparsers: subparsers object to extend with the subcommand
    :type subparsers: object returned by add_subparsers
    """
    parser_break = subparsers.add_parser("break", help="break a substitution cipher")
    parser_break.set_defaults(subparser=parser_break, func=command_break)
    args_add_language(parser_break)
    args_add_text_or_in_file(parser_break, "--ciphertext")
    parser_break.add_argument(
        "--consolidate",
        metavar="<int>",
        type=int,
        help=(
            "how often the same key must be found before it is regarded as the best "
            "solution. Default is 3. Lower values provide faster but unreliable "
            "results. If unsure don't touch it."
        ),
        default=3,
    )
    parser_break.add_argument(
        "--max-tries",
        metavar="<int>",
        type=int,
        help=(
            "the maximum number of hill climbings attempts. If no solution is found "
            "before this value is reached the best solution so far will be provided."
        ),
        default=1000,
    )


def args_decode(subparsers):
    """Defines the arguments for the subcommand "decode"

    :param subparsers: subparsers object to extend with the subcommand
    :type subparsers: object returned by add_subparsers
    """
    parser_decode = subparsers.add_parser(
        "decode", help="decode a substitution cipher with a given key"
    )
    parser_decode.set_defaults(subparser=parser_decode, func=command_decode)
    args_add_key_alternatives(parser_decode)
    args_add_alphabet(parser_decode)
    args_add_text_or_in_file(parser_decode, "--ciphertext")
    args_add_out_file(parser_decode, "--plaintext")


def args_encode(subparsers):
    """Defines the arguments for the subcommand "decode"

    :param subparsers: subparsers object to extend with the subcommand
    :type subparsers: object returned by add_subparsers
    """
    parser_encode = subparsers.add_parser(
        "encode", help="encode a plaintext with a given key"
    )
    parser_encode.set_defaults(subparser=parser_encode, func=command_encode)
    args_add_key_alternatives(parser_encode, random=True)
    args_add_alphabet(parser_encode)
    args_add_text_or_in_file(parser_encode, "--plaintext")
    args_add_out_file(parser_encode, "--ciphertext")


def args_fitness(subparsers):
    """Defines the arguments for the subcommand "fitness"

    :param subparsers: subparsers object to extend with the subcommand
    :type subparsers: object returned by add_subparsers
    """
    parser_fitness = subparsers.add_parser(
        "fitness", help="calculate the fitness for a given plaintext"
    )
    parser_fitness.set_defaults(subparser=parser_fitness, func=command_fitness)
    args_add_language(parser_fitness)
    args_add_text_or_in_file(parser_fitness, "--plaintext")


def args_quadgrams(subparsers):
    """Defines the arguments for the subcommand "quadgrams"

    :param subparsers: subparsers object to extend with the subcommand
    :type subparsers: object returned by add_subparsers
    """
    parser_quadgrams = subparsers.add_parser(
        "quadgrams", help="create quadgrams from a given text corpus"
    )
    parser_quadgrams.set_defaults(subparser=parser_quadgrams, func=command_quadgrams)
    args_add_alphabet(parser_quadgrams)
    parser_quadgrams.add_argument(
        "--corpus",
        metavar="<path>",
        help=(
            "The file name of the text corpus file. Only characters from the alphabet "
            "are considered (case insensitive). If this parameter is not provided,"
            "the text corpus will be read from STDIN."
        ),
        default=None,
    )
    parser_quadgrams.add_argument(
        "--quadgrams",
        metavar="<path>",
        help=(
            "The name of the file where the generated quadgrams are stored. "
            "The provided format is JSON. If this parameter is not provided, the "
            "quadgrams will be printed to STDOUT."
        ),
        default=None,
    )


def args_info(subparsers):
    """Defines the arguments for the subcommand "info"

    :param subparsers: subparsers object to extend with the subcommand
    :type subparsers: object returned by add_subparsers
    """
    parser_info = subparsers.add_parser(
        "info", help="print various information about the quadgram file"
    )
    parser_info.set_defaults(subparser=parser_info, func=command_info)
    args_add_language(parser_info)


def args_version(subparsers):
    """Defines the arguments for the subcommand "version"

    :param subparsers: subparsers object to extend with the subcommand
    :type subparsers: object returned by add_subparsers
    """
    parser_version = subparsers.add_parser(
        "version", help="print the version of the tool"
    )
    parser_version.set_defaults(subparser=parser_version, func=command_version)


def build_parser():
    """Creates the parser object

    :return: the parser object as created with argparse
    :rtype: :class:`argparse.ArgumentParser`
    """
    parser = argparse.ArgumentParser(
        description="A collection of tools to work with substitution ciphers"
    )
    subparsers = parser.add_subparsers(help="subcommands to execute", dest="command",)

    args_break(subparsers)
    args_decode(subparsers)
    args_encode(subparsers)
    args_fitness(subparsers)
    args_quadgrams(subparsers)
    args_info(subparsers)
    args_version(subparsers)
    return parser


def main():
    """The entry point for the command line interface
    """

    parser = build_parser()
    args = parser.parse_args()

    if args.command is None:
        parser.print_help()
    else:
        args.func(args)



# -*- coding: utf-8 -*-
import pytest
from subbreaker.breaker import Breaker
from subbreaker.key import AlphabetInvalid


def test_generate_quadgrams_invalid_alphabet(fixturefiles_dir, tmp_path):
    input_file = str(fixturefiles_dir / "quadgram_corpus.txt")
    output_file = str(tmp_path / "XXX.json")
    with open(input_file) as in_fh, open(output_file, "w") as out_fh:
        with pytest.raises(AlphabetInvalid):
            Breaker.generate_quadgrams(in_fh, out_fh, alphabet="ABCDEa")


def test_generate_quadgrams(fixturefiles_dir, tmp_path):
    input_file = str(fixturefiles_dir / "quadgram_corpus.txt")
    output_file = str(tmp_path / "XXX.json")
    with open(input_file) as in_fh, open(output_file, "w") as out_fh:
        Breaker.generate_quadgrams(in_fh, out_fh)

    with open(output_file) as quadgram_fh, open(input_file) as text_fh:
        breaker = Breaker(quadgram_fh)
        assert round(breaker.calc_fitness_file(text_fh)) == 100


def test_quadgram_fitness(quadgram_fh):
    breaker = Breaker(quadgram_fh)
    text1 = (
        "The museum will be a lasting physical testament to his hard work and "
        "vision, and will house the prestigious collection he cared so deeply about, "
        "for many years to come."
    )
    text2 = (
        "Heute ist jeder Autohersteller in der Lage starke Motoren zu bauen, doch "
        "alles hat seine Grenzen, sonst waeren ja alle anderen die sich ans Gesetz "
        "halten die Dummen."
    )
    text3 = (
        "Agl qrxlrq okii bl t itxakhj ugexknti alxatqlha ad gkx gtsm odsy thm "
        "pkxkdh, thm okii gdrxl agl uslxakjkdrx ndiilnakdh gl ntslm xd mlluie tbdra, "
        "vds qthe eltsx ad ndql."
    )
    fitness1 = breaker.calc_fitness(text1)
    fitness2 = breaker.calc_fitness(text2)
    fitness3 = breaker.calc_fitness(text3)
    assert 95 < fitness1 < 105
    assert fitness3 < fitness2 < fitness1
    assert 130 < breaker.calc_fitness("tion")


def test_alphabet_too_long(fixturefiles_dir, tmp_path):
    input_file = str(fixturefiles_dir / "quadgram_corpus.txt")
    output_file = str(tmp_path / "XXX.json")
    with pytest.raises(AlphabetInvalid, match=r"characters"):
        with open(input_file) as in_fh, open(output_file, "w") as out_fh:
            Breaker.generate_quadgrams(
                in_fh, out_fh, alphabet="abcdefghijklmnopqrstuvwxyz0123456789",
            )


def test_text_too_short(quadgram_fh):
    breaker = Breaker(quadgram_fh)
    with pytest.raises(ValueError, match=r"three characters"):
        breaker.calc_fitness("346523230 a b 47146")
    with pytest.raises(ValueError, match=r"three characters"):
        breaker.calc_fitness("346523230 a b c 47146")


def test_break_text(quadgram_fh, fixed_ciphertext):
    breaker = Breaker(quadgram_fh)
    result = breaker.break_cipher(fixed_ciphertext)
    assert result.key.startswith("wisdom")
    assert 95 < result.fitness < 105
    assert str(result).startswith("key = wisdom")
    assert breaker.info.alphabet == "abcdefghijklmnopqrstuvwxyz"
    assert breaker.info.nbr_quadgrams > 90000000
    assert breaker.info.most_frequent_quadgram == "tion"
    assert 0 < breaker.info.average_fitness < 50
    assert 120 < breaker.info.max_fitness < 150
    assert breaker.key._alphabet == "abcdefghijklmnopqrstuvwxyz"
    assert breaker.key._key.startswith("wisdom")


def test_breaker_text_too_short(quadgram_fh):
    breaker = Breaker(quadgram_fh)
    with pytest.raises(ValueError, match=r"ciphertext is too short"):
        breaker.break_cipher("34623230 a b  c 47146")
    result = breaker.break_cipher("34623230 a b  c d47146", max_rounds=1)
    assert result.fitness > 0


def test_break_max_tries_invalid(quadgram_fh, fixed_ciphertext):
    breaker = Breaker(quadgram_fh)
    with pytest.raises(
        ValueError, match=r"maximum number of rounds not in the valid range"
    ):
        breaker.break_cipher(fixed_ciphertext, max_rounds=10001)


def test_break_consolidate_invalid(quadgram_fh, fixed_ciphertext):
    breaker = Breaker(quadgram_fh)
    with pytest.raises(ValueError, match=r"consolidate parameter out of valid range"):
        breaker.break_cipher(fixed_ciphertext, consolidate=31)



# -*- coding: utf-8 -*-
import pytest
import io
from subbreaker.key import Key, AlphabetInvalid, KeyInvalid


def test_key_alphabet_lower():
    assert Key("abc", alphabet="ABc")._alphabet == "abc"


def test_key_double_char_in_alphabet():
    with pytest.raises(AlphabetInvalid, match=r"unique"):
        Key("abc", alphabet="aba")


def test_key_key_lower():
    assert Key("AbC", alphabet="abc")._key == "abc"


def test_key_double_char_in_key():
    with pytest.raises(KeyInvalid, match=r"unique"):
        Key("abca", alphabet="abcd")


def test_key_key_wrong_length():
    with pytest.raises(KeyInvalid, match=r"long"):
        Key("ab", alphabet="abc")
    with pytest.raises(KeyInvalid, match=r"long"):
        Key("abcd", alphabet="abc")


def test_key_key_incompatible_set():
    with pytest.raises(KeyInvalid, match=r"same set"):
        Key("abd", alphabet="abc")


def test_key_xcode():
    key = Key("abcdefghijklmnopqrstuvwxyz")
    assert key.decode("Hallo 23") == "Hallo 23"
    assert key.encode("Hallo 23") == "Hallo 23"


def test_key_xcode_file():
    key = Key("defghijklmnopqrstuvwxyzabc")
    output = io.StringIO()
    key.encode_file(plaintext_fh=io.StringIO("Hi there"), ciphertext_fh=output)
    assert output.getvalue() == "Kl wkhuh"
    output2 = io.StringIO()
    key.decode_file(ciphertext_fh=io.StringIO("Kl wkhuh"), plaintext_fh=output2)
    assert output2.getvalue() == "Hi there"



# -*- coding: utf-8 -*-
import sys
import re
import io
import json
import pytest

import subbreaker.command as command


def test_command_version(capsys):
    cmd = "subbreaker version"
    sys.argv = cmd.split()
    command.main()
    captured = capsys.readouterr()
    assert re.match(r"^\d+\.\d+\.\d+", captured.out)


def test_command_generate_quadgram(fixturefiles_dir, tmp_path):
    corpus = fixturefiles_dir / "quadgram_corpus.txt"
    quadgram_file = tmp_path / "YYY.json"
    cmd = "subbreaker quadgrams --corpus {} --quadgrams {}".format(
        str(corpus), str(quadgram_file)
    )
    sys.argv = cmd.split()
    command.main()
    with open(str(quadgram_file)) as fh:
        obj = json.load(fh)
    assert obj["alphabet"] == "abcdefghijklmnopqrstuvwxyz"
    assert type(obj["nbr_quadgrams"]) == int
    assert type(obj["most_frequent_quadgram"]) == str
    assert type(obj["max_fitness"]) == int
    assert type(obj["average_fitness"]) == float
    assert type(obj["quadgrams"]) == list
    assert type(obj["quadgrams"][0]) == int
    assert len(obj["quadgrams"]) == 32 * 32 * 32 * 32


def test_command_fitness_text(capsys):
    cmd = "subbreaker fitness --text Hello --lang EN"
    sys.argv = cmd.split()
    command.main()
    captured = capsys.readouterr()
    assert 20 < float(captured.out) < 120


def test_command_fitness_file(capsys, fixturefiles_dir):
    text_file = fixturefiles_dir / "quadgram_corpus.txt"
    cmd = "subbreaker fitness --plaintext {} --lang EN".format(str(text_file))
    sys.argv = cmd.split()
    command.main()
    captured = capsys.readouterr()
    assert 95 < float(captured.out) < 105


def test_command_decode_text(capsys):
    cmd = "subbreaker decode --text Khoor --key defghijklmnopqrstuvwxyzabc"
    sys.argv = cmd.split()
    command.main()
    captured = capsys.readouterr()
    assert captured.out.strip() == "Hello"


def test_command_decode_double_char_in_alphabet(capsys):
    cmd = "subbreaker decode --text Khoor --key abcde --alphabet abcdc"
    sys.argv = cmd.split()
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        command.main()
    captured = capsys.readouterr()
    assert captured.err.startswith("usage: subbreaker decode")
    assert "subbreaker decode: error:" in captured.err
    assert "unique" in captured.err
    assert pytest_wrapped_e.value.code == 2


def test_command_decode_double_char_in_key(capsys):
    cmd = "subbreaker decode --text Khoor --key abcdc --alphabet abcde"
    sys.argv = cmd.split()
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        command.main()
    captured = capsys.readouterr()
    assert captured.err.startswith("usage: subbreaker decode")
    assert "subbreaker decode: error:" in captured.err
    assert "unique" in captured.err
    assert pytest_wrapped_e.value.code == 2


def test_command_decode_alphabet_key_unequal_length(capsys):
    cmd = "subbreaker decode --text Khoor --key abcd --alphabet abcde"
    sys.argv = cmd.split()
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        command.main()
    captured = capsys.readouterr()
    assert captured.err.startswith("usage: subbreaker decode")
    assert "subbreaker decode: error:" in captured.err
    assert "long" in captured.err
    assert pytest_wrapped_e.value.code == 2


def test_command_decode_alphabet_key_incompatible_set(capsys):
    cmd = "subbreaker decode --text Khoor --key abcd --alphabet abce"
    sys.argv = cmd.split()
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        command.main()
    captured = capsys.readouterr()
    assert captured.err.startswith("usage: subbreaker decode")
    assert "subbreaker decode: error:" in captured.err
    assert "same set" in captured.err
    assert pytest_wrapped_e.value.code == 2


def test_command_decode_keyword(capsys):
    cmd = "subbreaker decode --text zebrascdfghijklmnopqtuvwxy --keyword Zebrasber"
    sys.argv = cmd.split()
    command.main()
    captured = capsys.readouterr()
    assert captured.out.strip() == "abcdefghijklmnopqrstuvwxyz"


def test_command_decode_file(capsys, hello_txt):
    cmd = "subbreaker decode --ciphertext {} --key defghijklmnopqrstuvwxyzabc".format(
        hello_txt
    )
    sys.argv = cmd.split()
    command.main()
    captured = capsys.readouterr()
    assert captured.out.strip() == "Ebiil"


def test_command_decode_no_args(capsys):
    sys.argv = "subbreaker decode".split()
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        command.main()
    captured = capsys.readouterr()
    assert captured.err.startswith("usage: subbreaker decode")
    assert pytest_wrapped_e.value.code == 2


def test_command_encode_text(capsys):
    cmd = "subbreaker encode --text Hello --key defghijklmnopqrstuvwxyzabc"
    sys.argv = cmd.split()
    command.main()
    captured = capsys.readouterr()
    assert captured.out.strip() == "Khoor"


def test_command_encode_file(capsys, hello_txt):
    cmd = "subbreaker encode --plaintext {} --key defghijklmnopqrstuvwxyzabc".format(
        hello_txt
    )
    sys.argv = cmd.split()
    command.main()
    captured = capsys.readouterr()
    assert captured.out.strip() == "Khoor"


def test_command_encode_no_args(capsys):
    sys.argv = "subbreaker encode".split()
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        command.main()
    captured = capsys.readouterr()
    assert "--key" in captured.err
    assert "--random" in captured.err
    assert pytest_wrapped_e.value.code == 2


def test_command_encode_double_char_in_alphabet(capsys):
    cmd = "subbreaker encode --text Khoor --key abcde --alphabet abcdc"
    sys.argv = cmd.split()
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        command.main()
    captured = capsys.readouterr()
    assert captured.err.startswith("usage: subbreaker encode")
    assert "subbreaker encode: error:" in captured.err
    assert "unique" in captured.err
    assert pytest_wrapped_e.value.code == 2


def test_command_encode_double_char_in_key(capsys):
    cmd = "subbreaker encode --text Khoor --key abcdc --alphabet abcde"
    sys.argv = cmd.split()
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        command.main()
    captured = capsys.readouterr()
    assert captured.err.startswith("usage: subbreaker encode")
    assert "subbreaker encode: error:" in captured.err
    assert "unique" in captured.err
    assert pytest_wrapped_e.value.code == 2


def test_command_encode_alphabet_key_unequal_length(capsys):
    cmd = "subbreaker encode --text Khoor --key abcd --alphabet abcde"
    sys.argv = cmd.split()
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        command.main()
    captured = capsys.readouterr()
    assert captured.err.startswith("usage: subbreaker encode")
    assert "subbreaker encode: error:" in captured.err
    assert "long" in captured.err
    assert pytest_wrapped_e.value.code == 2


def test_command_encode_alphabet_key_incompatible_set(capsys):
    cmd = "subbreaker encode --text Khoor --key abcd --alphabet abce"
    sys.argv = cmd.split()
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        command.main()
    captured = capsys.readouterr()
    assert captured.err.startswith("usage: subbreaker encode")
    assert "subbreaker encode: error:" in captured.err
    assert "same set" in captured.err
    assert pytest_wrapped_e.value.code == 2


def test_command_encode_random(capsys):
    cmd = "subbreaker encode --text Hello --random"
    sys.argv = cmd.split()
    command.main()
    captured = capsys.readouterr()
    key = captured.err.strip()
    ciphertext = captured.out.strip()
    cmd = "subbreaker decode --text {} --key {}".format(ciphertext, key)
    sys.argv = cmd.split()
    command.main()
    captured = capsys.readouterr()
    assert captured.out.strip() == "Hello"


def test_command_encode_keyword(capsys):
    cmd = "subbreaker encode --text abcdefghijklmnopqrstuvwxyz --keyword Zebrasber"
    sys.argv = cmd.split()
    command.main()
    captured = capsys.readouterr()
    assert captured.out.strip() == "zebrascdfghijklmnopqtuvwxy"


def test_command_break_text(capsys, fixed_ciphertext):
    cmd = "subbreaker break --text"
    sys.argv = cmd.split() + [fixed_ciphertext]
    command.main()
    captured = capsys.readouterr()
    assert "Alphabet: abcdefghijklmnopqrstuvwxyz" in captured.out
    assert "Key:      wisdom" in captured.out
    assert "Fitness: 10" in captured.out
    assert "The trouble with having an open mind," in captured.out


def test_command_break_file(capsys, ciphertext_file_name):
    cmd = "subbreaker break --ciphertext {}".format(ciphertext_file_name)
    sys.argv = cmd.split()
    command.main()
    captured = capsys.readouterr()
    assert "Alphabet: abcdefghijklmnopqrstuvwxyz" in captured.out
    assert "Key:      wisdom" in captured.out
    assert "Fitness: 10" in captured.out
    assert "The trouble with having an open mind," in captured.out


def test_command_break_invalid_max_tries(capsys):
    cmd = "subbreaker break --max-tries 10001"
    sys.argv = cmd.split()
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        command.main()
    captured = capsys.readouterr()
    assert captured.err.startswith("usage: subbreaker break")
    assert "subbreaker break: error:" in captured.err
    assert "must be in the range" in captured.err
    assert pytest_wrapped_e.value.code == 2


def test_command_break_invalid_consolidate(capsys):
    cmd = "subbreaker break --consolidate 31"
    sys.argv = cmd.split()
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        command.main()
    captured = capsys.readouterr()
    assert captured.err.startswith("usage: subbreaker break")
    assert "subbreaker break: error:" in captured.err
    assert "must be in the range" in captured.err
    assert pytest_wrapped_e.value.code == 2


def test_command_break_no_args(monkeypatch, fixed_ciphertext, capsys):
    monkeypatch.setattr("sys.stdin", io.StringIO(fixed_ciphertext))
    sys.argv = "subbreaker break".split()
    command.main()
    captured = capsys.readouterr()
    assert "Alphabet: abcdefghijklmnopqrstuvwxyz" in captured.out
    assert "Key:      wisdom" in captured.out
    assert "Fitness: 10" in captured.out
    assert "The trouble with having an open mind," in captured.out


def test_command_info(capsys):
    cmd = "subbreaker info"
    sys.argv = cmd.split()
    command.main()
    captured = capsys.readouterr()
    assert "Quadgram file:" in captured.out
    assert "subbreaker/quadgram/EN.json" in captured.out
    assert "Alphabet: abcdefghijklmnopqrstuvwxyz" in captured.out
    assert "Length of alphabet: 26" in captured.out
    assert "Number of quagrams:" in captured.out
    assert "Most frequent quadgram: tion" in captured.out
    assert "Fitness for most frequent quadgram: 13" in captured.out
    assert "Fitness for random text: 2" in captured.out


def test_no_command(capsys):
    cmd = "subbreaker"
    sys.argv = cmd.split()
    command.main()
    captured = capsys.readouterr()
    assert captured.out.startswith("usage: subbreaker [-h]")
    assert "A collection of tools" in captured.out



# -*- coding: utf-8 -*-
"""Basic functions for encoding/decoding substitution ciphers with a given key
"""
import sys


class AlphabetInvalid(Exception):
    """An exception raised when an alphabet is invalid"""
    pass


class KeyInvalid(Exception):
    """An exception raised when a key is invalid"""
    pass


class Key(object):
    """Uses a key and an alphabet for transcoding substitution ciphers.

    The first character of the alphabet corresponds to the first character of the
    key, the second character of the alphabet to the second character of the key,
    and so on. The alphabet can consist of any characters (including e.g.
    umlauts), and the length is variable, i.e., it is not restricted to the 26 letters
    of the alphabet.

    :example:
        ::

            Alphabet: abcdefghijklmnopqrstuvwxyz
            Key:      zebrascdfghijklmnopqtuvwxy

        The letter "a" from the plaintext maps to "z" in the ciphertext, "b" to "e",
        and so on. Thus the plaintext "flee at once. we are discovered!" is enciphered
        as "siaa zq lkba. va zoa rfpbluaoar!"

        This example was taken from
        `Wikipedia <https://en.wikipedia.org/wiki/Substitution_cipher>`_.

    :param str key: The key to use. Must have the same length than the alphabet.
        It is case insensitive.
    :param str alphabet: The set of characters which define the alphabet.
        Characters which are not in the alphabet will be ignored when transcoding.
    """

    _DEFAULT_ALPHABET = "abcdefghijklmnopqrstuvwxyz"

    def __init__(self, key, alphabet=_DEFAULT_ALPHABET):
        """Instantiate object
        """
        self._alphabet = self.check_alphabet(alphabet)
        self._key = self.check_key(key, self._alphabet)

        camel_key = self._upper(key) + key.lower()
        camel_alphabet = self._upper(alphabet) + alphabet.lower()
        self._encode = str.maketrans(camel_alphabet, camel_key)
        self._decode = str.maketrans(camel_key, camel_alphabet)

    @staticmethod
    def check_alphabet(alphabet):
        """Checks an alphabet for consistency

        Checks, if each character is unique.

        :param str alphabet: the alphabet to check
        :return: the alphabet in normalized form (i.e., in lower cases)
        :rtype: str
        :raises: AlphabetInvalid if the check fails
        """
        alphabet = alphabet.lower()
        if len(alphabet) != len(set(alphabet)):
            raise AlphabetInvalid("alphabet characters must be unique")
        return alphabet

    @staticmethod
    def check_key(key, alphabet):
        """Checks a key for consistency against a given alphabet

        It is assumed that the given alphabet has already been check for consistency
        before. The following checks are performed:

        - the characters in the key must be unique
        - the key must have the same length than the alphabet
        - the set of characters in the key must be the same than the set of characters
          in the alphabet

        :param str key: the key to be validated
        :param str alphabet: the alphabet against which the key is validated
        :return: the validated key in normalized form (i.e., in lower cases)
        :rtype: str
        :raises: KeyInvalid if one of the checks fails
        """
        key = key.lower()
        key_set = set(key)
        if len(key) != len(key_set):
            raise KeyInvalid("key characters must be unique")

        if len(key) != len(alphabet):
            raise KeyInvalid("key must be as long as the alphabet")
        if key_set != set(alphabet):
            raise KeyInvalid(
                "key must use the same set of characters than the alphabet"
            )
        return key

    @staticmethod
    def _upper(string):
        """
        Converts a string to upper case in a safe way

        Reason for this function is the German "ß".
        Problem: "ß".upper() results in "SS" which corrupts the xcoding translation
        table. Therefore in such a case the character is simply taken as it is and
        is not converted.

        :Example:
            "Viele Grüße".upper() results in "VIELE GRÜSSE"
            _upper("Viele Grüße") results in "VIELE GRÜßE"

        :param str string: the string to be converted to upper case
        :return: the string converted to upper case
        :rtype: str
        """
        return "".join(
            [char.upper() if len(char.upper()) == 1 else char for char in string]
        )

    def decode(self, ciphertext):
        """Decodes a ciphertext with the given key into the plaintext

        :param str ciphertext: the ciphertext to decode with the given key
        :return: the resulting plaintext
        :rtype: str
        """
        return ciphertext.translate(self._decode)

    def encode(self, plaintext):
        """Encodes a plaintext with the given key into the ciphertext

        :param str plaintext: the plaintext to encode with the given key
        :return: the resulting ciphertext
        :rtype: str
        """
        return plaintext.translate(self._encode)

    def decode_file(self, ciphertext_fh=sys.stdin, plaintext_fh=sys.stdout):
        """Decodes ciphertext read from the given file handle

        :param ciphertext_fh: the file handle (i.e., a read()-supporting file like
            object) the ciphertext is read from. Defaults to STDIN.
        :type ciphertext_fh: file object
        :param plaintext_fh: the file handle (i.e., a .write()-supporting file like
            object) the resulting plaintext is written to. Defaults to STDOUT.
        :type plaintext_fh: file object
        """
        for line in ciphertext_fh:
            plaintext_fh.write(self.decode(line))

    def encode_file(self, plaintext_fh=sys.stdin, ciphertext_fh=sys.stdout):
        """Encodes plaintext read from the given file handle

        :param plaintext_fh: the file handle (i.e., a read()-supporting file like
            object) the plaintext is read from. Defaults to STDIN.
        :type plaintext_fh: file object
        :param ciphertext_fh: the file handle (i.e., a .write()-supporting file like
            object) the resulting ciphertext is written to. Defaults to STDOUT.
        :type ciphertext_fh: file object
        """
        for line in plaintext_fh:
            ciphertext_fh.write(self.encode(line))
